'''
This file is part of RTL Multi-User Server, 
	that makes multi-user access to your DVB-T dongle used as an SDR.
Copyright (c) 2013 by Andras Retzler <retzlerandras@gmail.com>

RTL Multi-User Server is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

RTL Multi-User Server is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RTL Multi-User Server.  If not, see <http://www.gnu.org/licenses/>.
'''

import socket 
import sys
import array
import time
import logging
import os

print "RTL Multi-User Server v0.1a, made at HA5KFU Amateur Radio Club (http://ha5kfu.hu)"
print "    code by Andras Retzler, HA7ILM"
print "    distributed under GNU GPL v3"
print 

# === Load configuration script ===
if len(sys.argv)==1:
	print "Warning! Configuration script not specified. I will use: \"config_rtl.py\""
	config_script="config_rtl"
else:
	config_script=sys.argv[1]
cfg=__import__(config_script)

print "debug config allow gain set:", cfg.allow_gain_set

if cfg.setuid_on_start:
	os.setuid(cfg.uid)

def ip_match(this,ip_ranges,for_allow):
	if not len(ip_ranges):
		return 1 #empty list matches all ip addresses
	for ip_range in ip_ranges:
		print this[0:len(ip_range)], ip_range
		if this[0:len(ip_range)]==ip_range:
			return 1
	return 0

def ip_access_control(ip):
	if(not cfg.use_ip_access_control): return 1
	allowed=0
	if(cfg.order_allow_deny):
		if ip_match(ip,cfg.allowed_ip_ranges,1): allowed=1
		if ip_match(ip,cfg.denied_ip_ranges,0): allowed=0
	else:
		if ip_match(ip,cfg.denied_ip_ranges,0): 
			allowed=0
			print "deny"
		if ip_match(ip,cfg.allowed_ip_ranges,1): 
			allowed=1
			print "allow"
	return allowed
		

# === Set up logging ===
log = logging.getLogger("rtl_mus")
log.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.DEBUG)
stream_handler.setFormatter(formatter)
log.addHandler(stream_handler)
file_handler = logging.FileHandler(cfg.log_file_path)
file_handler.setLevel(logging.INFO)
file_handler.setFormatter(formatter)
log.addHandler(file_handler)
log.info("Server is UP")

server_missing_logged=0
rtl_dongle_identifier=''

def rtl_tcp_connect(close=0):
	global rtltcp_client_socket
	global server_missing_logged
	global rtl_dongle_identifier
	if not server_missing_logged:
		log.info("rtl_tcp host connection init: host=" + cfg.rtl_tcp_host + " port="+ str(cfg.rtl_tcp_port))
	if close:
		try:
			rtltcp_client_socket.close()
		except socket.error as err:
			log.error("rtl_tcp host connection close: "+str(err[0])+" "+err[1])
	rtltcp_client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	try:
		rtltcp_client_socket.connect((cfg.rtl_tcp_host, cfg.rtl_tcp_port))
		rtl_dongle_identifier=rtltcp_client_socket.recv(12)	
		#rtl_tcp server sends us a 12 byte info on the dongle used at server init
		#we'll send it to any clients connected to us
	except socket.error as err:
		if not server_missing_logged:
			log.error("rtl_tcp host connection: "+str(err[0])+" "+err[1])
			server_missing_logged=1
		return 0
	rtltcp_client_socket.setblocking(1)	# this does the trick, this is the ONLY blocking IO used
	rtltcp_client_socket.settimeout(3)
	log.info("rtl_tcp host connection estabilished")
	server_missing_logged=0
	return 1

rtl_tcp_connect()

my_listening_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
my_listening_socket.bind((cfg.my_ip, cfg.my_listening_port))
my_listening_socket.listen(5)
my_listening_socket.setblocking(0)
log.info("server listening on port: "+str(cfg.my_listening_port))

#I might implement decimation later, but not this way. - Andris, HA7ILM
'''
def average(input_data,average_items):
	if average_items==1 or len(input_data)==0:
		return input_data
	input_bytes=bytearray(input_data)
	output_bytes=bytearray()	
	j=0
	averaged_i=0
	averaged_q=0
	i_or_q_sample=0
	for i in range(0,len(input_data)):
		if i_or_q_sample:
			averaged_q+=input_bytes[i]
		else:
			averaged_i+=input_bytes[i]
		if i_or_q_sample==0:
			j+=1
		if j>=average_items:
			j=0
			output_bytes.append(averaged_i/average_items)
			output_bytes.append(averaged_q/average_items)
			averaged_i=0
			averaged_q=0
		i_or_q_sample=~i_or_q_sample
	return output_bytes
'''
	
def handle_command(command, client_id):
	param=array.array("I", command[1:5])[0]
	param=socket.ntohl(param)
	command_id=ord(command[0])
	client_info=str(client_id)+"@"+client_sockets[client_id][1][0]+":"+str(client_sockets[client_id][1][1])
	if(time.time()-client_start_times[client_id]<cfg.client_cant_set_until):
		log.info("deny: "+client_info+" -> client can't set anything until "+str(cfg.client_cant_set_until)+" seconds")
		return 0
	if command_id == 1:
		if max(map((lambda r: param>=r[0] and param<=r[1]),cfg.freq_allowed_ranges)):
			log.debug("allow: "+client_info+" -> set freq "+str(param))
			return 1
		else:
			log.debug("deny: "+client_info+" -> set freq - out of range: "+str(param))
	elif command_id == 2:
		log.debug("deny: "+client_info+" -> set sample rate: "+str(param))
		return 0 # ordinary clients are not allowed to do this
	elif command_id == 3:
		log.debug("deny/allow: "+client_info+" -> set gain mode: "+str(param))
		return cfg.allow_gain_set # for ordinary clients auto-gain will be set
	elif command_id == 4:
		log.debug("deny/allow: "+client_info+" -> set gain: "+str(param))
		return cfg.allow_gain_set  # for ordinary clients auto-gain will be set
	elif command_id == 5:
		log.debug("deny: "+client_info+" -> set freq correction: "+str(param))
		return 0  # ordinary clients are not allowed to do this
	elif command_id == 6:
		log.debug("deny/allow: set if stage gain")
		return cfg.allow_gain_set
	elif command_id == 7:
		log.debug("deny: set test mode")
		return 0
	elif command_id == 8:
		log.debug("deny/allow: set agc mode")
		return cfg.allow_gain_set
	elif command_id == 9:
		log.debug("deny: set direct sampling")
		return 0
	elif command_id == 10:
		log.debug("deny: set offset tuning")
		return 0
	elif command_id == 11:
		log.debug("deny: set rtl xtal")
		return 0
	elif command_id == 12:
		log.debug("deny: set tuner xtal")
		return 0
	elif command_id == 13:
		log.debug("deny/allow: set tuner gain by index")
		return cfg.allow_gain_set
	else:
		log.debug("deny: "+client_info+" sent an ivalid command: "+str(param))
	return 0

def close_clients():
	for which in client_indexes_to_close:
		try:	
			log.info("client disconnected: "+str(which)+"@"+client_sockets[which][1][0])
			client_sockets[which][0].close()
		except socket.error as err:
			log.error("can't close client socket: "+ str(err[0])+" "+err[1])
		del client_sockets[which]
		del client_waiting_data[which]
		del client_start_times[which]

client_sockets=[]
client_waiting_data=[]
client_start_times=[]
commands=''
command_send_errors=0

while 1:
	#First we try to ACCEPT new connections	
	try:
		temp_client_socket = my_listening_socket.accept() # we get [socket, address]
		temp_client_socket[0].setblocking(0)
		if(ip_access_control(temp_client_socket[1][0])):
			client_sockets.append(temp_client_socket)
			client_start_times.append(time.time())
			client_waiting_data.append([])
			temp_client_socket[0].send(rtl_dongle_identifier)
			log.info("client accepted: "+str(len(client_sockets)-1)+"@"+temp_client_socket[1][0]+":"+str(temp_client_socket[1][1])+"  users now: "+str(len(client_sockets)))
		else:
			log.info("client denied: "+str(len(client_sockets)-1)+"@"+temp_client_socket[1][0]+":"+str(temp_client_socket[1][1])+" blocked by ip")
			temp_client_socket[0].close()
	except socket.error as err:
		if err[0] != 11:
			log.error("client accept(): "+str(err[0])+" "+err[1])

	#We try to RECEIVE COMMANDS from our clients already connected
	for client_id in range(len(client_sockets)):
		client=client_sockets[client_id]
		try:		
			for i in range(1, 10): # we only take care of ten commands per loop from a client
				new_command = client[0].recv(5) # this is because sizeof(struct command) is 5 in rtl_tcp.
				if len(new_command)>=5:
					if handle_command(new_command, client_id):		
						commands+=new_command
		except socket.error as err:
			if err[0] != 11:
				log.error("client recv(): "+str(err[0])+" "+err[1])

	#We send the commands to the rtl_tcp server
	if len(commands)>0:
		try:
			rtltcp_client_socket.send(commands)
			command_send_errors=0
			commands=''
		except socket.error as err:
			command_send_errors+=1
			log.error("can't send command to RTL_TCP server, this happened",command_send_error,"times continously.")
		except msg:
			log.error("rtl_tcp send(): "+msg) # might remove this	

	#We get data from the rtl_tcp server	
	new_data=''
	try:
		while len(new_data)<16348:
			new_data+=rtltcp_client_socket.recv(1024)	# should do LPF and decimation here, if required for slower download speeds
	except socket.error as err:
		if err[0] != 11:
			log.error("rtl_tcp recv(): "+str(err[0])+" "+err[1])
			while not rtl_tcp_connect(1):
				time.sleep(1)
	if(len(new_data)==0): continue #!
	# print len(new_data), "bytes of data acquired from RTL_TCP server"

	#We send data to all client sockets
	client_indexes_to_close=[]
	for client_i in range(0,len(client_sockets)):		
		client=client_sockets[client_i]		
		try:
			waiting_data_len=0
			if len(client_waiting_data[client_i]):
				waiting_data_len=reduce((lambda x,y:x+y),map(len,client_waiting_data[client_i]))
			if len(new_data)+waiting_data_len<cfg.buffer_size:
				client_waiting_data[client_i].append(new_data) 
				#http://itamarst.org/writings/pycon05/fast.html "Bad String Concatenation" was a cool tip
			else:
				log.error("client cache full: "+str(client_i)+"@"+client[1][0]+" now clearing cache")
				client_waiting_data[client_i]=[new_data]
				# print waiting_data_len, len(client_waiting_data[client_i])
				# client_indexes_to_close.append(client_i) # We rather clear cache and hope client to resync soon
			while len(client_waiting_data[client_i]):
				client[0].send(client_waiting_data[client_i][0]) # FIXME might be problems here if we do it in the wrong order
				del client_waiting_data[client_i][0]
			#print "info: client", client_i, "got", len(client_waiting_data[client_i]), "bytes"
		except socket.error as err:
			if(err[0]==32): # broken pipe
				client_indexes_to_close.append(client_i)	
			if(err[0]!=11):		
				log.error("client send(): "+str(err[0])+" "+err[1])
	close_clients()

rtltcp_client_socket.close()
